import click

from crunchy_bot.logging.logger import Logger, LogTag, LogColor


class ClickLogger(Logger):
    """
    Logger that uses click logging utilities to log to stdout.
    Format in the form of:
        [ TAG ] MESSAGE
    """

    def success(self, message: str):
        self.log(LogTag.SUCCESS, message, LogColor.SUCCESS)

    def info(self, message: str):
        self.log(LogTag.INFO, message, LogColor.INFO)

    def warn(self, message: str):
        self.log(LogTag.WARN, message, LogColor.WARN)

    def error(self, message: str):
        self.log(LogTag.ERROR, message, LogColor.ERROR)

    def critical(self, message: str):
        self.log(LogTag.CRITICAL, message, LogColor.CRITICAL)

    def log(self, tag: LogTag, message: str, color: LogColor):
        click.secho(f"[ {tag.name[:4]} ] {message}", fg=color.value)



import os
from pathlib import Path
from unittest import mock
from unittest.mock import mock_open, patch

import pytest

from crunchy_bot.config.config_parser import ConfigParser, Config
from crunchy_bot.config.json_config_parser import JsonConfigParser
from crunchy_bot.logging.logger import Logger


@pytest.fixture()
def logger() -> Logger:
    return mock.create_autospec(Logger)


@pytest.fixture()
def json_config_parser(logger: Logger) -> ConfigParser:
    return JsonConfigParser(logger=logger)


@pytest.fixture()
def config_data() -> dict:
    return {
        "crunchy_username": "crunchy_user",
        "crunchy_password": "crunchy_pass",
        "reddit_client_id": "client_id",
        "reddit_client_secret": "client_secret",
        "reddit_user_agent": "CrunchyBot:v4.0.0 (hosted by /u/{YOUR_USERNAME})",
        "reddit_username": "reddit_user",
        "reddit_password": "reddit_pass",
        "log_dir": "/tmp/crunchybot/logs",
    }


def test_parse(logger: Logger, json_config_parser: ConfigParser, config_data: dict):
    with patch("crunchy_bot.config.json_config_parser.open", mock_open()) as open_mock:
        with patch("crunchy_bot.config.json_config_parser.json") as json_mock:
            json_mock.load.return_value = config_data

            config_path = Path.home()
            config = json_config_parser.parse(config_path)

            logger.info.assert_any_call("Fetching Account Data...")
            open_mock.assert_any_call(config_path.resolve().as_posix(), "r")
            json_mock.load.assert_called()
            logger.info.assert_any_call("Fetched Account Data")

            assert config == Config(**config_data)


def test_parse_file_not_found_exception(
    logger: Logger, json_config_parser: ConfigParser, config_data: dict
):
    with patch("crunchy_bot.config.json_config_parser.open", mock_open()) as open_mock:
        with patch("crunchy_bot.config.json_config_parser.json") as json_mock:
            open_mock.side_effect = FileNotFoundError

            for k, v in config_data.items():
                os.environ[k.upper()] = v

            config_path = Path.home()
            config = json_config_parser.parse(config_path)

            logger.info.assert_any_call("Fetching Account Data...")
            open_mock.assert_any_call(config_path.resolve().as_posix(), "r")
            logger.info.assert_any_call("Fetched Account Data")
            json_mock.load.assert_not_called()

            assert config == Config(**config_data)



from enum import Enum


class LogTag(Enum):
    SUCCESS = 0
    INFO = 1
    WARN = 2
    ERROR = 3
    CRITICAL = 4


class LogColor(Enum):
    SUCCESS = "green"
    INFO = "blue"
    WARN = "yellow"
    ERROR = "red"
    CRITICAL = "pink"
    LOG = "white"


class Logger(object):
    def success(self, message: str):
        """
        Success level logging

        :param message: message to log
        """
        raise NotImplementedError("success is not implemented")

    def info(self, message: str):
        """
        Info level logging

        :param message: message to log
        """
        raise NotImplementedError("info is not implemented")

    def warn(self, message: str):
        """
        Warn level logging

        :param message: message to log
        """
        raise NotImplementedError("warn is not implemented")

    def error(self, message: str):
        """
        Error level logging

        :param message: message to log
        """
        raise NotImplementedError("error is not implemented")

    def critical(self, message: str):
        """
        Critical level logging

        :param message: message to log
        """
        raise NotImplementedError("critical is not implemented")

    def log(self, tag: LogTag, message: str, color: LogColor):
        """
        Custom level logging

        :param message: message to log
        """
        raise NotImplementedError("log is not implemented")



import json
import os
import pathlib
from typing import Optional

from crunchy_bot.config.config_parser import ConfigParser, Config
from crunchy_bot.logging.logger import Logger
from crunchy_bot.logging.noop_logger import NoopLogger


class JsonConfigParser(ConfigParser):
    def __init__(self, logger: Optional[Logger] = None):
        self.logger = logger if logger is not None else NoopLogger()

    def parse(self, path: pathlib.Path) -> Config:
        self.logger.info("Fetching Account Data...")
        posix_path = path.resolve().as_posix()

        data_dictionary = {}
        if path.exists():
            try:
                with open(posix_path, "r") as data_file:
                    data_dictionary = json.load(data_file)
            except FileNotFoundError as e:
                self.logger.warn(f"Unable to read {posix_path}: {str(e)}")
                self.logger.info("Attempting to load from environment")

        crunchy_username = data_dictionary.get(
            "crunchy_username", os.environ.get("CRUNCHY_USERNAME")
        )
        if crunchy_username is None:
            raise ValueError(
                f"Missing crunchy_username in {posix_path} and environment"
            )

        crunchy_password = data_dictionary.get(
            "crunchy_password", os.environ.get("CRUNCHY_PASSWORD")
        )
        if crunchy_password is None:
            raise ValueError(
                f"Missing crunchy_password in {posix_path} and environment"
            )

        reddit_client_id = data_dictionary.get(
            "reddit_client_id", os.environ.get("REDDIT_CLIENT_ID")
        )
        if reddit_client_id is None:
            raise ValueError(
                f"Missing reddit_client_id in {posix_path} and environment"
            )

        reddit_client_secret = data_dictionary.get(
            "reddit_client_secret", os.environ.get("REDDIT_CLIENT_SECRET")
        )
        if reddit_client_secret is None:
            raise ValueError(
                f"Missing reddit_client_secret in {posix_path} and environment"
            )

        reddit_user_agent = data_dictionary.get(
            "reddit_user_agent", os.environ.get("REDDIT_USER_AGENT")
        )
        if reddit_user_agent is None:
            raise ValueError(
                f"Missing reddit_user_agent in {posix_path} and environment"
            )

        reddit_username = data_dictionary.get(
            "reddit_username", os.environ.get("REDDIT_USERNAME")
        )
        if reddit_username is None:
            raise ValueError(f"Missing reddit_username in {posix_path} and environment")

        reddit_password = data_dictionary.get(
            "reddit_password", os.environ.get("REDDIT_PASSWORD")
        )
        if reddit_password is None:
            raise ValueError(f"Missing reddit_password in {posix_path} and environment")

        log_dir = data_dictionary.get("log_dir", os.environ.get("LOG_DIR"))
        if log_dir is None:
            raise ValueError(f"Missing log_dir in {posix_path}")

        self.logger.info("Fetched Account Data")

        return Config(
            crunchy_username=crunchy_username,
            crunchy_password=crunchy_password,
            reddit_client_id=reddit_client_id,
            reddit_client_secret=reddit_client_secret,
            reddit_user_agent=reddit_user_agent,
            reddit_username=reddit_username,
            reddit_password=reddit_password,
            log_dir=log_dir,
        )



import pathlib
from typing import Any
from unittest import mock
from unittest.mock import patch

import pytest
from selenium.webdriver.common.keys import Keys

from crunchy_bot.config.config_parser import Config
from crunchy_bot.fetcher.guest_pass_fetcher import GuestPassFetcher
from crunchy_bot.logging.logger import Logger


@pytest.fixture()
def config() -> Config:
    return Config(
        crunchy_username="username",
        crunchy_password="password",
        reddit_client_id="",
        reddit_client_secret="",
        reddit_user_agent="",
        reddit_username="",
        reddit_password="",
        log_dir="/path/to/logdir",
    )


@pytest.fixture()
def logger() -> Logger:
    return mock.create_autospec(Logger)


class Row(object):
    def __init__(
        self,
        created_at: str,
        guest_pass: str,
        description: str,
        status: str,
        expiration: str,
        redeemer: str,
        action: str,
    ):
        self.cells = [
            Cell(created_at),
            Cell(guest_pass),
            Cell(description),
            Cell(status),
            Cell(expiration),
            Cell(redeemer),
            Cell(action),
        ]


class Cell(object):
    def __init__(self, text: str):
        self.text = text

    def __repr__(self):
        return f"Cell text={self.text}"


@patch("crunchy_bot.fetcher.guest_pass_fetcher.WebDriverWait", autospec=True)
@patch("crunchy_bot.fetcher.guest_pass_fetcher.webdriver", autospec=True)
@patch("crunchy_bot.fetcher.guest_pass_fetcher.Options", autospec=True)
@patch("crunchy_bot.fetcher.guest_pass_fetcher.pathlib", autospec=True)
def test_fetch(
    pathlib_mock, options_mock, webdriver_mock, webdriver_wait_mock, config, logger
):
    path_mock = mock.create_autospec(pathlib.Path)
    pathlib_mock.Path.return_value = path_mock
    path_mock.exists.return_value = True
    path_mock.joinpath.return_value = path_mock
    log_path = f"{config.log_dir}/chrome.log"
    path_mock.as_posix.return_value = log_path
    driver_mock = webdriver_mock.Chrome()
    table_mock = mock.MagicMock()
    guest_pass_tables = [table_mock]
    driver_mock.find_elements_by_class_name.return_value = guest_pass_tables
    valid_row_mock = mock.MagicMock()
    invalid_row_mock = mock.MagicMock()
    table_mock.find_elements_by_tag_name.return_value = [
        valid_row_mock,
        invalid_row_mock,
    ]
    valid_guest_pass = "abcd123"
    valid_row = Row(
        created_at="1/1/2000",
        guest_pass=valid_guest_pass,
        description="description",
        status="Valid",
        expiration="1/2/2000",
        redeemer="",
        action="",
    )
    valid_row_mock.find_elements_by_tag_name.return_value = valid_row.cells
    invalid_row = Row(
        created_at="12/30/1999",
        guest_pass="invalid",
        description="description",
        status="Expired",
        expiration="12/31/1999",
        redeemer="",
        action="",
    )
    invalid_row_mock.find_elements_by_tag_name.return_value = invalid_row.cells

    fetcher = GuestPassFetcher(config, logger=logger)
    assert fetcher.fetch() == [valid_guest_pass]

    logger.info.assert_any_call("Fetching Guest Passes...")
    pathlib_mock.Path.assert_any_call(config.log_dir)
    path_mock.mkdir.assert_not_called()
    options_mock.assert_any_call()
    options_mock().add_argument.assert_any_call(f"--log-path={log_path}")
    options_mock().add_argument.assert_any_call("--headless")
    webdriver_mock.Chrome.assert_any_call(options=options_mock())
    driver_mock.get.assert_any_call("https://www.crunchyroll.com/login?next=%2F")
    webdriver_wait_mock.assert_any_call(driver_mock, 20)
    # EC does not implement __eq__ or __hash__
    webdriver_wait_mock(driver_mock, 20).until.assert_called()
    webdriver_wait_mock(driver_mock, 20).until(Any).send_keys.assert_any_call(
        config.crunchy_username
    )
    webdriver_wait_mock(driver_mock, 20).until(Any).send_keys.assert_any_call(
        config.crunchy_password
    )
    webdriver_wait_mock(driver_mock, 20).until(Any).send_keys.assert_any_call(
        Keys.ENTER
    )
    driver_mock.quit.assert_not_called()
    driver_mock.get.assert_any_call(
        "https://www.crunchyroll.com/acct/?action=guestpass"
    )
    driver_mock.close.assert_any_call()
    logger.info.assert_any_call("Fetched Guest Passes")



from typing import Sequence


class Fetcher(object):
    def fetch(self, debug: bool = False) -> Sequence[str]:
        """
        Return a sequence of Guest Pass strings.

        :param debug: Enable debugging mode. It is up to the implementor to define what this means.
        :return: Sequence of Guest Pass strings.
        """
        raise NotImplementedError("fetch is not implemented")
